@@ -10,9 +10,16 @@ require "labels.pl";
 my $key;
 # sort so that we get a consistent assignment for nodes with multiple labels 
 foreach $label (sort keys %external_labels) {
-  $key = $external_labels{$label};
-  $key =~ s|^/||;
-  $nodes{$key} = $label;
+  #
+  # If the label can't be used as a filename on non-Unix platforms,
+  # skip it.  Such labels may be used internally within the documentation,
+  # but will never be used for filename generation.
+  #
+  if ($label =~ /^([-.a-zA-Z0-9]+)$/) {
+    $key = $external_labels{$label};
+    $key =~ s|^/||;
+    $nodes{$key} = $label;
+  }
 }
 
 # This adds the "internal" labels added for indexing.  These labels will not
@@ -32,7 +39,7 @@ foreach $label (keys %internal_labels) {
 while (<>) {
   # don't want to do one s/// per line per node
   # so look for lines with hrefs, then do s/// on nodes present
-  if (/(HREF|href)=[\"\']([^\#\"\']*)html[\#\"\']/) {
+  if (/(HREF|href)=[\"\']node\d+\.html[\#\"\']/) {
     @parts = split(/(HREF|href)\=[\"\']/);
     shift @parts;
     for $node (@parts) {
