@@ -5,9 +5,9 @@ for the Distutils compiler abstraction model."""
 
 # created 1999/07/05, Greg Ward
 
-__revision__ = "$Id: ccompiler.py,v 1.35 2000/09/27 02:24:21 gward Exp $"
+__revision__ = "$Id: ccompiler.py,v 1.41 2001/12/06 20:51:35 fdrake Exp $"
 
-import sys, os
+import sys, os, re
 from types import *
 from copy import copy
 from distutils.errors import *
@@ -61,7 +61,7 @@ class CCompiler:
     #     different versions of libfoo.a in different locations.  I
     #     think this is useless without the ability to null out the
     #     library search path anyways.
-    
+
 
     # Subclasses that rely on the standard filename generation methods
     # implemented below should override these; see the comment near
@@ -159,7 +159,7 @@ class CCompiler:
             setattr(self, key, split_quoted(value))
         else:
             setattr(self, key, value)
-        
+
 
 
     def _find_macro (self, name):
@@ -352,7 +352,7 @@ class CCompiler:
         else:
             raise TypeError, \
                   "'include_dirs' (if supplied) must be a list of strings"
-                    
+
         return (output_dir, macros, include_dirs)
 
     # _fix_compile_args ()
@@ -364,8 +364,9 @@ class CCompiler:
         list of all object files and a dictionary telling which source
         files can be skipped.
         """
-        # Get the list of expected output (object) files 
+        # Get the list of expected output (object) files
         objects = self.object_filenames (sources,
+                                         strip_dir=1,
                                          output_dir=output_dir)
 
         if self.force:
@@ -400,7 +401,7 @@ class CCompiler:
             raise TypeError, \
                   "'objects' must be a list or tuple of strings"
         objects = list (objects)
-            
+
         if output_dir is None:
             output_dir = self.output_dir
         elif type (output_dir) is not StringType:
@@ -559,7 +560,7 @@ class CCompiler:
         Raises LibError on failure.
         """
         pass
-    
+
 
     # values for target_desc parameter in link()
     SHARED_OBJECT = "shared_object"
@@ -620,7 +621,7 @@ class CCompiler:
         """
         raise NotImplementedError
 
-    
+
     # Old 'link_*()' methods, rewritten to use the new 'link()' method.
 
     def link_shared_lib (self,
@@ -635,13 +636,13 @@ class CCompiler:
                          extra_preargs=None,
                          extra_postargs=None,
                          build_temp=None):
-        self.link(CCompiler.SHARED_LIBRARY, objects, 
+        self.link(CCompiler.SHARED_LIBRARY, objects,
                   self.library_filename(output_libname, lib_type='shared'),
                   output_dir,
                   libraries, library_dirs, runtime_library_dirs,
                   export_symbols, debug,
                   extra_preargs, extra_postargs, build_temp)
-    
+
 
     def link_shared_object (self,
                             objects,
@@ -672,9 +673,9 @@ class CCompiler:
                          debug=0,
                          extra_preargs=None,
                          extra_postargs=None):
-        self.link(CCompiler.EXECUTABLE, objects, 
+        self.link(CCompiler.EXECUTABLE, objects,
                   self.executable_filename(output_progname), output_dir,
-                  libraries, library_dirs, runtime_library_dirs, None, 
+                  libraries, library_dirs, runtime_library_dirs, None,
                   debug, extra_preargs, extra_postargs, None)
 
 
@@ -791,8 +792,8 @@ class CCompiler:
                           output_dir=''):
 
         if output_dir is None: output_dir = ''
-        if lib_type not in ("static","shared"):
-            raise ValueError, "'lib_type' must be \"static\" or \"shared\""
+        if lib_type not in ("static","shared","dylib"):
+            raise ValueError, "'lib_type' must be \"static\", \"shared\" or \"dylib\""
         fmt = getattr (self, lib_type + "_lib_format")
         ext = getattr (self, lib_type + "_lib_extension")
 
@@ -834,11 +835,47 @@ class CCompiler:
 # class CCompiler
 
 
-# Map a platform ('posix', 'nt') to the default compiler type for
-# that platform.
-default_compiler = { 'posix': 'unix',
-                     'nt': 'msvc',
-                   }
+# Map a sys.platform/os.name ('posix', 'nt') to the default compiler
+# type for that platform. Keys are interpreted as re match
+# patterns. Order is important; platform mappings are preferred over
+# OS names.
+_default_compilers = (
+
+    # Platform string mappings
+
+    # on a cygwin built python we can use gcc like an ordinary UNIXish
+    # compiler
+    ('cygwin.*', 'unix'),
+
+    # OS name mappings
+    ('posix', 'unix'),
+    ('nt', 'msvc'),
+    ('mac', 'mwerks'),
+
+    )
+
+def get_default_compiler(osname=None, platform=None):
+
+    """ Determine the default compiler to use for the given platform.
+
+        osname should be one of the standard Python OS names (i.e. the
+        ones returned by os.name) and platform the common value
+        returned by sys.platform for the platform in question.
+
+        The default values are os.name and sys.platform in case the
+        parameters are not given.
+
+    """
+    if osname is None:
+        osname = os.name
+    if platform is None:
+        platform = sys.platform
+    for pattern, compiler in _default_compilers:
+        if re.match(pattern, platform) is not None or \
+           re.match(pattern, osname) is not None:
+            return compiler
+    # Default to Unix compiler
+    return 'unix'
 
 # Map compiler types to (module_name, class_name) pairs -- ie. where to
 # find the code that implements an interface to this compiler.  (The module
@@ -853,6 +890,8 @@ compiler_class = { 'unix':    ('unixccompiler', 'UnixCCompiler',
                                "Mingw32 port of GNU C Compiler for Win32"),
                    'bcpp':    ('bcppcompiler', 'BCPPCompiler',
                                "Borland C++ Compiler"),
+                   'mwerks':  ('mwerkscompiler', 'MWerksCompiler',
+                               "MetroWerks CodeWarrior"),
                  }
 
 def show_compilers():
@@ -862,7 +901,7 @@ def show_compilers():
     # XXX this "knows" that the compiler option it's describing is
     # "--compiler", which just happens to be the case for the three
     # commands that use it.
-    from distutils.fancy_getopt import FancyGetopt 
+    from distutils.fancy_getopt import FancyGetopt
     compilers = []
     for compiler in compiler_class.keys():
         compilers.append(("compiler="+compiler, None,
@@ -870,7 +909,7 @@ def show_compilers():
     compilers.sort()
     pretty_printer = FancyGetopt(compilers)
     pretty_printer.print_help("List of available compilers:")
-    
+
 
 def new_compiler (plat=None,
                   compiler=None,
@@ -892,15 +931,15 @@ def new_compiler (plat=None,
 
     try:
         if compiler is None:
-            compiler = default_compiler[plat]
-        
+            compiler = get_default_compiler(plat)
+
         (module_name, class_name, long_description) = compiler_class[compiler]
     except KeyError:
         msg = "don't know how to compile C/C++ code on platform '%s'" % plat
         if compiler is not None:
             msg = msg + " with '%s' compiler" % compiler
         raise DistutilsPlatformError, msg
-              
+
     try:
         module_name = "distutils." + module_name
         __import__ (module_name)
