@@ -278,6 +278,9 @@ Exception__str__(PyObject *self, PyObject *args)
 	    out = NULL;
 	break;
     }
+    case -1:
+        PyErr_Clear();
+        /* Fall through */
     default:
         out = PyObject_Str(args);
         break;
@@ -409,6 +412,9 @@ SystemExit__init__(PyObject *self, PyObject *args)
     case 1:
         code = PySequence_GetItem(args, 0);
         break;
+    case -1:
+        PyErr_Clear();
+        /* Fall through */
     default:
         Py_INCREF(args);
         code = args;
@@ -515,6 +521,10 @@ EnvironmentError__init__(PyObject *self, PyObject *args)
 	    goto finally;
 	}
 	break;
+
+    case -1:
+	PyErr_Clear();
+	break;
     }
 
     Py_INCREF(Py_None);
