@@ -233,10 +233,8 @@ if missing: raise "Missing Types"
 		self.includepath = [':', INCLUDEDIR]
 
 	def initpatterns(self):
-#		self.head_pat = "^extern pascal[ \t]+" # XXX Mac specific!
 		self.head_pat = "^EXTERN_API[^_]"
 		self.tail_pat = "[;={}]"
-#		self.type_pat = "pascal[ \t\n]+\(<type>[a-zA-Z0-9_ \t]*[a-zA-Z0-9_]\)[ \t\n]+"
 		self.type_pat = "EXTERN_API" + \
 						"[ \t\n]*([ \t\n]*" + \
 						"\(<type>[a-zA-Z0-9_* \t]*[a-zA-Z0-9_*]\)" + \
@@ -244,14 +242,12 @@ if missing: raise "Missing Types"
 		self.name_pat = "\(<name>[a-zA-Z0-9_]+\)[ \t\n]*"
 		self.args_pat = "(\(<args>\([^(;=)]+\|([^(;=)]*)\)*\))"
 		self.whole_pat = self.type_pat + self.name_pat + self.args_pat
-#		self.sym_pat = "^[ \t]*\(<name>[a-zA-Z0-9_]+\)[ \t]*=" + \
-#		               "[ \t]*\(<defn>[-0-9'\"(][^\t\n,;}]*\),?"
 		self.sym_pat = "^[ \t]*\(<name>[a-zA-Z0-9_]+\)[ \t]*=" + \
 		               "[ \t]*\(<defn>[-0-9_a-zA-Z'\"(][^\t\n,;}]*\),?"
-		self.asplit_pat = "^\(<type>.*[^a-zA-Z0-9_]\)\(<name>[a-zA-Z0-9_]+\)$"
+		self.asplit_pat = "^\(<type>.*[^a-zA-Z0-9_]\)\(<name>[a-zA-Z0-9_]+\)\(<array>\[\]\)?$"
 		self.comment1_pat = "\(<rest>.*\)//.*"
 		# note that the next pattern only removes comments that are wholly within one line
-		self.comment2_pat = "\(<rest>.*\)/\*.*\*/"
+		self.comment2_pat = "\(<rest1>.*\)/\*.*\*/\(<rest2>.*\)"
 
 	def compilepatterns(self):
 		for name in dir(self):
@@ -328,6 +324,16 @@ if missing: raise "Missing Types"
 		return file
 
 	def setinput(self, scan = sys.stdin):
+		if not type(scan) in (TupleType, ListType):
+			scan = [scan]
+		self.allscaninputs = scan
+		self._nextinput()
+		
+	def _nextinput(self):
+		if not self.allscaninputs:
+			return 0
+		scan = self.allscaninputs[0]
+		self.allscaninputs = self.allscaninputs[1:]
 		self.closescan()
 		if scan:
 			if type(scan) == StringType:
@@ -339,6 +345,7 @@ if missing: raise "Missing Types"
 			self.scanfile = file
 			self.scanmine = mine
 		self.lineno = 0
+		return 1
 
 	def openinput(self, filename):
 		if not os.path.isabs(filename):
@@ -360,6 +367,8 @@ if missing: raise "Missing Types"
 			raise Error, "input file not set"
 		self.line = self.scanfile.readline()
 		if not self.line:
+			if self._nextinput():
+				return self.getline()
 			raise EOFError
 		self.lineno = self.lineno + 1
 		return self.line
@@ -379,7 +388,7 @@ if missing: raise "Missing Types"
 			self.report("(No symbol definitions will be written)")
 		else:
 			self.report("defsfile = %s", `self.defsfile.name`)
-			self.defsfile.write("# Generated from %s\n\n" % `inputname`)
+			self.defsfile.write("# Generated from %s\n\n" % `os.path.split(inputname)[1]`)
 			self.writeinitialdefs()
 		self.alreadydone = []
 		try:
@@ -388,8 +397,8 @@ if missing: raise "Missing Types"
 				except EOFError: break
 				if self.comment1.match(line) >= 0:
 					line = self.comment1.group('rest')
-				if self.comment2.match(line) >= 0:
-					line = self.comment2.group('rest')
+				while self.comment2.match(line) >= 0:
+					line = self.comment2.group('rest1')+self.comment2.group('rest2')
 				if self.defsfile and self.sym.match(line) >= 0:
 					self.dosymdef()
 					continue
@@ -412,6 +421,10 @@ if missing: raise "Missing Types"
 		raw = self.line
 		while self.tail.search(raw) < 0:
 			line = self.getline()
+			if self.comment1.match(line) >= 0:
+				line = self.comment1.group('rest')
+			while self.comment2.match(line) >= 0:
+				line = self.comment2.group('rest1')+self.comment2.group('rest2')
 			raw = raw + line
 		self.processrawspec(raw)
 
@@ -458,7 +471,10 @@ if missing: raise "Missing Types"
 		if self.asplit.match(part) < 0:
 			self.error("Indecipherable argument: %s", `part`)
 			return ("unknown", part, mode)
-		type, name = self.asplit.group('type', 'name')
+		type, name, array = self.asplit.group('type', 'name', 'array')
+		if array:
+			# array matches an optional [] after the argument name
+			type = type + " ptr "
 		type = regsub.gsub("\*", " ptr ", type)
 		type = string.strip(type)
 		type = regsub.gsub("[ \t]+", "_", type)
@@ -560,15 +576,23 @@ class Scanner_PreUH3(Scanner):
 	def initpatterns(self):
 		Scanner.initpatterns(self)
 		self.head_pat = "^extern pascal[ \t]+" # XXX Mac specific!
-		self.tail_pat = "[;={}]"
 		self.type_pat = "pascal[ \t\n]+\(<type>[a-zA-Z0-9_ \t]*[a-zA-Z0-9_]\)[ \t\n]+"
-		self.name_pat = "\(<name>[a-zA-Z0-9_]+\)[ \t\n]*"
-		self.args_pat = "(\(<args>\([^(;=)]+\|([^(;=)]*)\)*\))"
 		self.whole_pat = self.type_pat + self.name_pat + self.args_pat
 		self.sym_pat = "^[ \t]*\(<name>[a-zA-Z0-9_]+\)[ \t]*=" + \
 		               "[ \t]*\(<defn>[-0-9'\"][^\t\n,;}]*\),?"
-		self.asplit_pat = "^\(<type>.*[^a-zA-Z0-9_]\)\(<name>[a-zA-Z0-9_]+\)$"
 
+class Scanner_OSX(Scanner):
+	"""Scanner for modern (post UH3.3) Universal Headers """
+	def initpatterns(self):
+		Scanner.initpatterns(self)
+		self.head_pat = "^EXTERN_API\(_C\)?"
+		self.type_pat = "EXTERN_API\(_C\)?" + \
+						"[ \t\n]*([ \t\n]*" + \
+						"\(<type>[a-zA-Z0-9_* \t]*[a-zA-Z0-9_*]\)" + \
+						"[ \t\n]*)[ \t\n]*"
+		self.whole_pat = self.type_pat + self.name_pat + self.args_pat
+		self.sym_pat = "^[ \t]*\(<name>[a-zA-Z0-9_]+\)[ \t]*=" + \
+		               "[ \t]*\(<defn>[-0-9_a-zA-Z'\"(][^\t\n,;}]*\),?"
 	
 def test():
 	input = "D:Development:THINK C:Mac #includes:Apple #includes:AppleEvents.h"
