@@ -51,6 +51,9 @@ class ScriptBinding:
     ]
 
     def __init__(self, editwin):
+        if not editwin.runnable:
+            self.menudefs = []
+            self.keydefs = {}
         self.editwin = editwin
         # Provide instance variables referenced by Debugger
         # XXX This should be done differently
